import numpy as np

from mdt.components_loader import bind_function
from mot.model_building.evaluation_models import GaussianEvaluationModel
from mdt.models.single import DMRISingleModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class S0TM(DMRISingleModelConfig):

    name = 'S0-TM'
    description = 'Full STEAM model for T1 decay by variation in mixing time. T2 contained in S0'
    model_expression = 'S0 * ExpT1DecTM'


class S0T2Steam(DMRISingleModelConfig):

    name = 'S0-T2steam'
    description = 'Full STEAM model for T2 decay by variation in echo time. T1 contained in S0'
    model_expression = 'S0 * ExpT2DecSTEAM'


class S0T2(DMRISingleModelConfig):

    name = 'S0-T2'
    description = 'Models the unweighted signal (aka. b0) with an extra T2.'
    model_expression = 'S0 * ExpT2Dec'

    # for proper initialization, please take the highest S0 value in your data.
    #inits = {'S0.s0': 50.0}
    upper_bounds = {'ExpT2Dec.T2': 0.15}
    #                'S0.s0': 150}


class S0T2Linear(DMRISingleModelConfig):

    description = 'Models the unweighted signal (aka. b0) with an extra T2.'
    model_expression = 'S0 + LinT2Dec'
    upper_bounds = {'LinT2Dec.R2': 1000,
                    'S0.s0': 7.0}

    @bind_function
    def _transform_observations(self, observations):
        return np.log(observations)


class S0_IRT1(DMRISingleModelConfig):

    name = 'S0-ExpT1DecIR'
    description = 'Model with multi-IR data (?)'
    model_expression = 'S0 * ExpT1DecIR'

    # for proper initialization, please take the highest S0 value in your data.
    inits = {'S0.s0': 50.0}
    upper_bounds = {'ExpT1DecIR.T2': 0.10,
                    'S0.s0': 150}


class S0T1GRE(DMRISingleModelConfig):

    name = 'S0-T1GRE'
    description = 'Models the unweighted signal (aka. b0) with an extra T1.'
    model_expression = 'S0 * ExpT1DecGRE'

    # for proper initialization, please take the highest S0 value in your data.
    inits = {'S0.s0': 50.0}
    upper_bounds = {'ExpT1DecGRE.T1': 1.0,
                    'S0.s0': 150}


class S0T2T2(DMRISingleModelConfig):

    name = 'S0-T2T2'
    description = 'Model for the unweighted signal with two T2 models, one for short T2 and one for long T2.'

    model_expression = '''
            S0 * ( (Weight(w_long) * ExpT2Dec(T2_long)) +
                   (Weight(w_short) * ExpT2Dec(T2_short))
                 )
        '''

    fixes = {'T2_long.T2': 0.5}
    upper_bounds = {'T2_short.T2': 0.08}

    post_optimization_modifiers = (
        ('T2_short.T2Weighted', lambda d: d['w_short.w'] * d['T2_short.T2']),
        ('T2_long.T2Weighted', lambda d: d['w_long.w'] * d['T2_long.T2']),
        ('T2.T2', lambda d: d['T2_short.T2Weighted'] + d['T2_long.T2Weighted'])
    )


class GRE_Relax(DMRISingleModelConfig):

    name = 'GRE_Relax'
    description = 'Model for estimating T1 and T2 from GRE data with variable TE, TR and flip angle.'
    model_expression = 'S0 * ExpT1ExpT2GRE'
    inits = {'ExpT1ExpT2GRE.T1': 0.2,
             'ExpT1ExpT2GRE.T2': 0.05,
             'S0.s0': 1e3}
    upper_bounds = {'ExpT1ExpT2GRE.T1': 1,
                    'ExpT1ExpT2GRE.T2': 0.5,
                    'S0.s0': 1e4}


class rSTEAM(DMRISingleModelConfig):

    name = 'rSTEAM'
    description = 'Model for estimating T1 and T2 from data with a variable TM and TE.'
    model_expression = 'S0 * ExpT1ExpT2STEAM'
    inits = {'ExpT1ExpT2STEAM.T2': 0.03,
             'ExpT1ExpT2STEAM.T1': 0.15}
    upper_bounds = {'ExpT1ExpT2STEAM.T1': 0.5,
                    'ExpT1ExpT2STEAM.T2': 0.1}


class S0GRELinear(DMRISingleModelConfig):
    """S0 is not the "real" s0 of the data, it is s0*(1 - exp(-TR / T1)).

    The real s0 can be only calculated AFTER T1 estimation.
    """
    name = 'S0-GRE-Linear'
    description = 'Model for estimating T1 of GRE data using B1+ map and several FA variations.'
    model_expression = 'S0 + LinT1GRE'
    evaluation_model = GaussianEvaluationModel()


class MPM(DMRISingleModelConfig):

    name = 'MPM'
    description = 'Model for estimating biological microstructure of the tissue/sample.'
    model_expression = 'S0 * MPM_Fit'
    upper_bounds = {'MPM_Fit.T1': 0.8}
    evaluation_model = GaussianEvaluationModel()


class LinMPM(DMRISingleModelConfig):

    name = 'LinMPM'
    description = 'Linear model for estimating biological microstructure of the tissue/sample.'
    model_expression = 'S0 + LinMPM_Fit'
    upper_bounds = {'LinMPM_Fit.T1': 0.8}
    evaluation_model = GaussianEvaluationModel()

    @bind_function
    def _transform_observations(self, observations):
        return np.log(observations)



__author__ = 'Robbert Harms'
__date__ = "2016-06-25"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class ProtocolIOError(Exception):
    """Custom exception class for protocol input output errors.

    This can be raised if a protocol is inconsistent or incomplete. It should not be raised for general IO errors,
    use the IO exception for that.
    """


class InsufficientProtocolError(Exception):
    """Indicates that the protocol constains insufficient information for fitting a specific model.

    This can be raised if a model misses a column it needs in the protocol, or if there are not enough shells, or...
    """


class NoiseStdEstimationNotPossible(Exception):
    """An exception that can be raised by any ComplexNoiseStdEstimator.

    This indicates that the noise std can not be estimated by the estimation routine.
    """



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
import argparse
import os
import textwrap

from argcomplete.completers import FilesCompleter

from mdt import init_user_settings
from mdt.gui.model_fit.qt_main import start_gui
from mdt.shell_utils import BasicShellApplication, get_citation_message

__author__ = 'Robbert Harms'
__date__ = "2015-08-18"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class GUI(BasicShellApplication):

    def __init__(self):
        init_user_settings(pass_if_exists=True)

    def _get_arg_parser(self):
        description = textwrap.dedent("""
            Launches the MDT Graphical User Interface.
        """)
        description += get_citation_message()

        parser = argparse.ArgumentParser(description=description, formatter_class=argparse.RawTextHelpFormatter)
        parser.add_argument('-d', '--dir', metavar='dir', type=str, help='the base directory for the file choosers',
                            default=None).completer = FilesCompleter()
        return parser

    def run(self, args):
        if args.dir:
            cwd = os.path.realpath(args.dir)
        else:
            cwd = os.getcwd()

        start_gui(cwd)

if __name__ == '__main__':
    GUI().start()



from mdt.models.single import DMRISingleModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallStick(DMRISingleModelConfig):

    ex_vivo_suitable = False
    description = 'The default Ball & Stick model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_stick) * Stick) )
    '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick.d': 1.7e-9}
    post_optimization_modifiers = [('FS', lambda results: 1 - results['w_ball.w'])]


class BallStickT2(BallStick):

    name = 'BallStick-T2'
    description = 'The Ball & Stick model with extra T2 weighting'
    model_expression = '''
        S0 * ExpT2Dec * ( (Weight(w_ball) * Ball) +
                          (Weight(w_stick) * Stick) )
    '''


class BallStickT2T2(BallStick):

    name = 'BallStick-T2T2'
    description = 'The Ball & Stick model with two extra T2 models'
    model_expression = '''
            S0 * ( (ExpT2Dec(T2_long) * Weight(w_ball) * Ball) +
                   (ExpT2Dec(T2_short) * Weight(w_stick) * Stick) )
        '''


class BallStickExVivo(BallStick):

    name = 'BallStick-ExVivo'
    in_vivo_suitable = False
    ex_vivo_suitable = True
    description = 'The Ball & Stick model with ex vivo defaults',
    fixes = {'Ball.d': 2.0e-9,
             'Stick.d': 0.6e-9}


class StickExVivo(DMRISingleModelConfig):

    name = 'Stick-ExVivo'
    in_vivo_suitable = False
    ex_vivo_suitable = True
    description = 'The Stick model with ex vivo defaults',
    fixes = {'Stick.d': 0.6e-9}
    model_expression = '''
        S0 * Stick
    '''


class BallStickT2ExVivo(BallStickExVivo):

    name = 'BallStick-T2-ExVivo'
    description = 'The Ball & Stick model with extra T2 weighting and exvivo defaults'
    model_expression = '''
            S0 * ExpT2Dec * ( (Weight(w_ball) * Ball) +
                              (Weight(w_stick) * Stick) )
        '''


class StickStickExVivo(DMRISingleModelConfig):

    name = 'StickStick-ExVivo'
    in_vivo_suitable = False
    ex_vivo_suitable = True
    description = 'The 2x Stick model ex vivo defaults'
    model_expression = '''
            S0 * ( (Weight(w_stick0) * Stick(Stick0)) +
                   (Weight(w_stick1) * Stick(Stick1)) )
        '''
    fixes = {'Stick0.d': 1.7e-9,
             'Stick1.d': 1.7e-9}


class BallStickStick(DMRISingleModelConfig):

    ex_vivo_suitable = False
    description = 'The Ball & 2x Stick model'
    model_expression = '''
            S0 * ( (Weight(w_ball) * Ball) +
                   (Weight(w_stick0) * Stick(Stick0)) +
                   (Weight(w_stick1) * Stick(Stick1)) )
        '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick0.d': 1.7e-9,
             'Stick1.d': 1.7e-9}
    post_optimization_modifiers = [('FS', lambda results: 1 - results['w_ball.w'])]


class BallStickStickExVivo(BallStickStick):

    name = 'BallStickStick-ExVivo'
    in_vivo_suitable = False
    ex_vivo_suitable = True
    description = 'The Ball & 2x Stick model with ex vivo defaults'
    fixes = {'Ball.d': 2.0e-9,
             'Stick0.d': 0.6e-9,
             'Stick1.d': 0.6e-9}


class BallStickStickStick(DMRISingleModelConfig):

    ex_vivo_suitable = False
    description = 'The Ball & 3x Stick model'
    model_expression = '''
            S0 * ( (Weight(w_ball) * Ball) +
                   (Weight(w_stick0) * Stick(Stick0)) +
                   (Weight(w_stick1) * Stick(Stick1)) +
                   (Weight(w_stick2) * Stick(Stick2)) )
        '''
    fixes = {'Ball.d': 3.0e-9,
             'Stick0.d': 1.7e-9,
             'Stick1.d': 1.7e-9,
             'Stick2.d': 1.7e-9}
    inits = {'w_stick2.w': 0}

    post_optimization_modifiers = [('FS', lambda results: 1 - results['w_ball.w'])]


class BallStickStickStickExVivo(BallStickStickStick):

    name = 'BallStickStickStick-ExVivo'
    ex_vivo_suitable = True
    in_vivo_suitable = False
    description = 'The Ball & 3x Stick model with ex vivo defaults'
    fixes = {'Ball.d': 2.0e-9,
             'Stick0.d': 0.6e-9,
             'Stick1.d': 0.6e-9,
             'Stick2.d': 0.6e-9}



from mdt.models.cascade import CascadeConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class BallStick(CascadeConfig):

    name = 'BallStick (Cascade)'
    description = 'Cascade for Ballstick'
    models = ('S0',
              'BallStick')


class BallStickS0(BallStick):

    name = 'BallStick (Cascade|S0)'


class BallStickExVivo(BallStick):

    name = 'BallStick-ExVivo (Cascade)'
    description = 'Cascade for Ballstick with ex vivo defaults.'
    models = ('S0',
              'BallStick-ExVivo')


class BallStickStick(CascadeConfig):

    name = 'BallStickStick (Cascade)'
    description = 'Cascade for BallStickStick.'
    models = ('BallStick (Cascade)',
              'BallStickStick')
    inits = {'BallStickStick': [('Stick0.theta', 'Stick.theta'),
                                ('Stick0.phi', 'Stick.phi'),
                                ('w_stick0.w', 'w_stick.w'),
                                ('w_stick1.w', 0.0)]}


class BallStickStickExVivo(BallStickStick):

    name = 'BallStickStick-ExVivo (Cascade)'
    description = 'Cascade for BallStickStick with ex vivo defaults.'
    models = ('BallStick-ExVivo (Cascade)',
              'BallStickStick-ExVivo')


class BallStickStickStick(CascadeConfig):

    name = 'BallStickStickStick (Cascade)'
    description = 'Cascade for BallStickStickStick.'
    models = ('BallStickStick (Cascade)',
              'BallStickStickStick')
    inits = {'BallStickStickStick': [('w_stick2.w', 0.0)]}


class BallStickStickStickExVivo(BallStickStickStick):

    name = 'BallStickStickStick-ExVivo (Cascade)'
    description = 'Cascade for BallStickStickStick with ex vivo defaults.'
    models = ('BallStickStick-ExVivo (Cascade)',
              'BallStickStickStick-ExVivo')



from mdt.models.single import DMRISingleModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class SSFP_BallStick(DMRISingleModelConfig):

    name = 'SSFP_BallStick'
    ex_vivo_suitable = False
    description = 'The SSFP Ball & Stick model'
    model_expression = '''
        S0 * ( (Weight(w_ball) * Ball) +
               (Weight(w_stick) * SSFP_Stick) )
    '''
    fixes = {'Ball.d': 3.0e-9,
             'SSFP_Stick.d': 1.7e-9}
    post_optimization_modifiers = [('FS', lambda results: 1 - results['w_ball.w'])]



from mdt.models.single import DMRISingleModelConfig
from mot.model_building.parameter_functions.dependencies import SimpleAssignment, AbstractParameterDependency

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class NoddiTortuosityParameterDependency(AbstractParameterDependency):

    def __init__(self, d, w_ec, w_ic, ):
        self._d = d
        self._w_ec = w_ec
        self._w_ic = w_ic

    @property
    def pre_transform_code(self):
        return '''
            mot_float_type _tortuosity_mult_{d} = {w_ec} / ({w_ec} + {w_ic});
            if(!isnormal(_tortuosity_mult_{d})){{
                _tortuosity_mult_{d} = 0.01;
            }}
        '''.format(d=self._d, w_ec=self._w_ec, w_ic=self._w_ic)

    @property
    def assignment_code(self):
        return '{d} * _tortuosity_mult_{d}'.format(d=self._d)

    @property
    def fixed(self):
        return True

    @property
    def has_side_effects(self):
        return False


class Noddi(DMRISingleModelConfig):

    ex_vivo_suitable = False
    description = 'The standard Noddi (NODDI) model'

    model_expression = '''
        S0 * ((Weight(w_csf) * Ball) +
              (Weight(w_ic) * Noddi_IC) +
              (Weight(w_ec) * Noddi_EC)
              )
    '''

    fixes = {'Noddi_IC.d': 1.7e-9,
             'Noddi_IC.R': 0.0,
             'Noddi_EC.d': 1.7e-9,
             'Ball.d': 3.0e-9}

    dependencies = (
        ('Noddi_EC.dperp0', NoddiTortuosityParameterDependency('Noddi_EC.d', 'w_ec.w', 'w_ic.w')),
        ('Noddi_EC.kappa', SimpleAssignment('Noddi_IC.kappa')),
        ('Noddi_EC.theta', SimpleAssignment('Noddi_IC.theta')),
        ('Noddi_EC.phi', SimpleAssignment('Noddi_IC.phi'))
    )

    post_optimization_modifiers = [
        ('NDI', lambda d: d['w_ic.w'] / (d['w_ic.w'] + d['w_ec.w'])),
        ('ODI', lambda d: d['Noddi_IC.odi'])
    ]


class Noddi2(DMRISingleModelConfig):

    ex_vivo_suitable = False
    description = 'The Noddi model with two IC and EC compartments'

    model_expression = '''
            S0 * (
                  (Weight(w_ic0) * Noddi_IC(Noddi_IC0)) +
                  (Weight(w_ec0) * Noddi_EC(Noddi_EC0)) +

                  (Weight(w_ic1) * Noddi_IC(Noddi_IC1)) +
                  (Weight(w_ec1) * Noddi_EC(Noddi_EC1)) +

                  (Weight(w_csf) * Ball)
            )
        '''

    fixes = {'Noddi_IC0.d': 1.7e-9,
             'Noddi_IC0.R': 0.0,
             'Noddi_IC1.d': 1.7e-9,
             'Noddi_IC1.R': 0.0,
             'Noddi_EC0.d': 1.7e-9,
             'Noddi_EC1.d': 1.7e-9,
             'Ball.d': 3.0e-9}

    dependencies = (
        ('Noddi_EC0.dperp0', NoddiTortuosityParameterDependency('Noddi_EC0.d', 'w_ec0.w', 'w_ic0.w')),
        ('Noddi_IC0.kappa', SimpleAssignment('((w_ic0.w + w_ec0.w) >= {cutoff}) * Noddi_IC0.kappa'.format(cutoff=0.01),
                                             fixed=False)),
        ('Noddi_EC0.kappa', SimpleAssignment('Noddi_IC0.kappa')),
        ('Noddi_EC0.theta', SimpleAssignment('Noddi_IC0.theta')),
        ('Noddi_EC0.phi', SimpleAssignment('Noddi_IC0.phi')),

        ('Noddi_EC1.dperp0', NoddiTortuosityParameterDependency('Noddi_EC1.d', 'w_ec1.w', 'w_ic1.w')),
        ('Noddi_IC1.kappa', SimpleAssignment('((w_ic1.w + w_ec1.w) >= {cutoff}) * Noddi_IC1.kappa'.format(cutoff=0.01),
                                             fixed=False)),
        ('Noddi_EC1.kappa', SimpleAssignment('Noddi_IC1.kappa')),
        ('Noddi_EC1.theta', SimpleAssignment('Noddi_IC1.theta')),
        ('Noddi_EC1.phi', SimpleAssignment('Noddi_IC1.phi')),
    )

    post_optimization_modifiers = (
        ('NDI0', lambda d: d['w_ic0.w'] / (d['w_ic0.w'] + d['w_ec0.w'])),
        ('ODI0', lambda d: d['Noddi_IC0.odi']),
        ('NDI1', lambda d: d['w_ic1.w'] / (d['w_ic1.w'] + d['w_ec1.w'])),
        ('ODI1', lambda d: d['Noddi_IC1.odi'])
    )



from mdt.models.cascade import CascadeConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CharmedR1(CascadeConfig):

    name = 'Charmed_r1 (Cascade)'
    description = 'Initializes the directions to Ball & Stick.'
    models = ('BallStick (Cascade)',
              'Charmed_r1')
    inits = {'Charmed_r1': [('CharmedRestricted0.theta', 'Stick.theta'),
                            ('CharmedRestricted0.phi', 'Stick.phi'),
                            ('Tensor.theta', 'Stick.theta'),
                            ('Tensor.phi', 'Stick.phi'),
                            ('w_res0.w', 'w_stick.w')]}


class CharmedR1S0(CascadeConfig):

    name = 'Charmed_r1 (Cascade|S0)'
    description = 'Cascade for Charmed r1 initialized with only an S0 fit.'
    models = ('S0',
              'Charmed_r1')


class CharmedR1Fixed(CascadeConfig):

    name = 'Charmed_r1 (Cascade|fixed)'
    description = 'Fixes the directions to Ball & Stick.'
    models = ('BallStick (Cascade)',
              'Charmed_r1')
    inits = {'Charmed_r1': [('Tensor.theta', 'Stick.theta'),
                            ('Tensor.phi', 'Stick.phi'),
                            ('w_res0.w', 'w_stick.w')]}
    fixes = {'Charmed_r1': [('CharmedRestricted0.theta', 'Stick.theta'),
                            ('CharmedRestricted0.phi', 'Stick.phi')]}


class CharmedR2(CascadeConfig):

    name = 'Charmed_r2 (Cascade)'
    description = 'Initializes the directions to 2x Ball & Stick.'
    models = ('BallStickStick (Cascade)',
              'Charmed_r2')
    inits = {'Charmed_r2': [('Tensor.theta', 'Stick0.theta'),
                            ('Tensor.phi', 'Stick0.phi'),
                            ('CharmedRestricted0.theta', 'Stick0.theta'),
                            ('CharmedRestricted0.phi', 'Stick0.phi'),
                            ('CharmedRestricted1.theta', 'Stick1.theta'),
                            ('CharmedRestricted1.phi', 'Stick1.phi'),
                            ('w_res0.w', 'w_stick0.w'),
                            ('w_res1.w', 'w_stick1.w')]}


class CharmedR2S0(CascadeConfig):

    name = 'Charmed_r2 (Cascade|S0)'
    description = 'Initializes with only an S0 fit.'
    models = ('S0',
              'Charmed_r2')


class CharmedR2Fixed(CascadeConfig):

    name = 'Charmed_r2 (Cascade|fixed)'
    description = 'Fixes the directions to 2x Ball & Stick.'
    models = ('BallStickStick (Cascade)',
              'Charmed_r2')
    inits = {'Charmed_r2': [('Tensor.theta', 'Stick0.theta'),
                            ('Tensor.phi', 'Stick0.phi'),
                            ('w_res0.w', 'w_stick0.w'),
                            ('w_res1.w', 'w_stick1.w')]}
    fixes = {'Charmed_r2': [('CharmedRestricted0.theta', 'Stick0.theta'),
                            ('CharmedRestricted0.phi', 'Stick0.phi'),
                            ('CharmedRestricted1.theta', 'Stick1.theta'),
                            ('CharmedRestricted1.phi', 'Stick1.phi'),
                            ]}


class Charmed(CascadeConfig):

    name = 'Charmed (Cascade)'
    description = 'Initializes the directions to 3x Ball & Stick.'
    models = ('BallStickStickStick (Cascade)',
              'Charmed')
    inits = {'Charmed': [('Tensor.theta', 'Stick0.theta'),
                         ('Tensor.phi', 'Stick0.phi'),
                         ('w_res0.w', 'w_stick0.w'),
                         ('w_res1.w', 'w_stick1.w'),
                         ('w_res2.w', 'w_stick2.w'),
                         ('CharmedRestricted0.theta', 'Stick0.theta'),
                         ('CharmedRestricted0.phi', 'Stick0.phi'),
                         ('CharmedRestricted1.theta', 'Stick1.theta'),
                         ('CharmedRestricted1.phi', 'Stick1.phi'),
                         ('CharmedRestricted2.theta', 'Stick2.theta'),
                         ('CharmedRestricted2.phi', 'Stick2.phi'),
                         ]}


class CharmedS0(CascadeConfig):

    name = 'Charmed (Cascade|S0)'
    description = 'Initializes with only an S0 fit.'
    models = ('S0',
              'Charmed')


class CharmedFixed(CascadeConfig):

    name = 'Charmed (Cascade|fixed)'
    description = 'Fixes the directions to 3x Ball & Stick.'
    models = ('BallStickStickStick (Cascade)',
              'Charmed')
    inits = {'Charmed': [('Tensor.theta', 'Stick0.theta'),
                         ('Tensor.phi', 'Stick0.phi'),
                         ('w_res0.w', 'w_stick0.w'),
                         ('w_res1.w', 'w_stick1.w'),
                         ('w_res2.w', 'w_stick2.w')]}
    fixes = {'Charmed': [('CharmedRestricted0.theta', 'Stick0.theta'),
                         ('CharmedRestricted0.phi', 'Stick0.phi'),
                         ('CharmedRestricted1.theta', 'Stick1.theta'),
                         ('CharmedRestricted1.phi', 'Stick1.phi'),
                         ('CharmedRestricted2.theta', 'Stick2.theta'),
                         ('CharmedRestricted2.phi', 'Stick2.phi')]}



VERSION = '0.8.15'
VERSION_STATUS = ''

_items = VERSION.split('-')
VERSION_NUMBER_PARTS = tuple(int(i) for i in _items[0].split('.'))
if len(_items) > 1:
    VERSION_STATUS = _items[1]
__version__ = VERSION



from mdt.models.compartments import CompartmentConfig
from mdt.components_loader import LibraryFunctionsLoader, bind_function

lib_loader = LibraryFunctionsLoader()

class SSFP_Stick(CompartmentConfig):

    parameter_list = ('g', 'd', 'TR', 'flip_angle', 'b1map', 'T1map', 'T2map')
    dependency_list = (lib_loader.load('MRIConstants'),)

    @bind_function
    def get_extra_results_maps(self, results_dict):
        return self._get_vector_result_maps(results_dict[self.name + '.theta'],
                                            results_dict[self.name + '.phi'])


import glob
import os
import mdt
from mdt.batch_utils import SimpleBatchProfile, BatchFitProtocolLoader, SimpleSubjectInfo

__author__ = 'Robbert Harms'
__date__ = "2015-07-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"

meta_info = {'title': 'HCP WU-Minn',
             'description': 'The profile for the WU-Minn data from the Human Connectome project',
             'directory_layout':
'''
This assumes that you downloaded and extracted the WU-Minn data in one folder which gives one folder per subject.

You can provide the noise standard deviation to use using a noise_std file containing a single float.

Example directory layout:
    /*/T1w/Diffusion/data.nii.gz
    /*/T1w/Diffusion/bvals
    /*/T1w/Diffusion/bvecs
    /*/T1w/Diffusion/nodif_brain_mask.nii.gz
    /*/T1w/Diffusion/grad_dev.nii.gz

Optional items (these will take precedence if present):
    /*/T1w/Diffusion/data.bval
    /*/T1w/Diffusion/data.bvec
    /*/T1w/Diffusion/data.prtcl
    /*/T1w/Diffusion/data_mask.nii(.gz)
    /*/T1w/Diffusion/noise_std
'''}


class HCP_WUMINN(SimpleBatchProfile):

    def __init__(self):
        super(HCP_WUMINN, self).__init__()
        self.use_gradient_deviations = False
        self._output_base_dir = 'T1w/Diffusion/output'
        self._append_mask_name_to_output_sub_dir = False
        self.models_to_fit = ('BallStick (Cascade)',
                              'Tensor (Cascade)',
                              'Noddi (Cascade|fixed)',
                              'Charmed_r1 (Cascade|fixed)',
                              'Charmed_r2 (Cascade|fixed)',
                              'Charmed (Cascade|fixed)')

    def _get_subjects(self):
        dirs = sorted([os.path.basename(f) for f in glob.glob(os.path.join(self._root_dir, '*'))])
        subjects = []
        for subject_id in dirs:
            pjoin = mdt.make_path_joiner(self._root_dir, subject_id, 'T1w', 'Diffusion')
            if os.path.isdir(pjoin()):
                dwi_fname = list(glob.glob(pjoin('data.nii*')))[0]

                noise_std = self._autoload_noise_std(subject_id, file_path=pjoin('noise_std'))

                bval_fname = pjoin('bvals')
                if os.path.isfile(pjoin('data.bval')):
                    bval_fname = pjoin('data.bval')

                bvec_fname = pjoin('bvecs')
                if os.path.isfile(pjoin('data.bvec')):
                    bvec_fname = pjoin('data.bvec')

                prtcl_fname = None
                if os.path.isfile(pjoin('data.prtcl')):
                    prtcl_fname = pjoin('data.prtcl')

                mask_fname = list(glob.glob(pjoin('nodif_brain_mask.nii*')))[0]
                if list(glob.glob(pjoin('data_mask.nii*'))):
                    mask_fname = list(glob.glob(pjoin('data_mask.nii*')))[0]

                if self.use_gradient_deviations:
                    grad_dev = pjoin('grad_dev.nii.gz')
                else:
                    grad_dev = None

                protocol_loader = BatchFitProtocolLoader(
                    pjoin(),
                    protocol_fname=prtcl_fname, bvec_fname=bvec_fname,
                    bval_fname=bval_fname, protocol_options={'Delta': 43.1e-3, 'delta': 10.6e-3,
                                                             'TE': 89.5e-3, 'TR': 5520e-3})

                output_dir = self._get_subject_output_dir(subject_id, mask_fname)

                subjects.append(SimpleSubjectInfo(subject_id, dwi_fname, protocol_loader, mask_fname,
                                                  output_dir, gradient_deviations=grad_dev, noise_std=noise_std))
        return subjects

    def __str__(self):
        return meta_info['title']



import numpy as np

from mdt.components_loader import bind_function
from mdt.models.compartments import CompartmentConfig
from mdt.cl_routines.mapping.dti_measures import DTIMeasures
from mdt.utils import eigen_vectors_from_tensor

__author__ = 'Robbert Harms'
__date__ = "2015-06-21"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(CompartmentConfig):

    parameter_list = ('g', 'b', 'd', 'dperp0', 'dperp1', 'theta', 'phi', 'psi')

    @bind_function
    def get_extra_results_maps(self, results_dict):
        eigen_vectors = eigen_vectors_from_tensor(results_dict[self.name + '.theta'], results_dict[self.name + '.phi'],
                                                  results_dict[self.name + '.psi'])

        eigen_values = np.squeeze(np.concatenate([m[..., None] for m in [results_dict[self.name + '.d'],
                                                                         results_dict[self.name + '.dperp0'],
                                                                         results_dict[self.name + '.dperp1']]], axis=1))

        if len(eigen_values.shape) < 2:
            # in the case of a single voxel, the eigen values collapses to a 1d array, but we need it to be 2d
            eigen_values = eigen_values[None, :]

        ranking = np.squeeze(np.argsort(eigen_values, axis=1)[:, ::-1])

        if len(ranking.shape) < 2:
            ranking = ranking[None, :]

        voxels_listing = np.arange(ranking.shape[0])
        sorted_eigen_values = [eigen_values[voxels_listing, ranking[:, ind]] for ind in range(ranking.shape[1])]

        extra_maps = {self.name + '.eigen_ranking': ranking}
        for ind in range(3):
            extra_maps.update({self.name + '.vec' + repr(ind): eigen_vectors[:, ind, :]})

            for dimension in range(3):
                extra_maps.update({self.name + '.vec' + repr(ind) + '_' + repr(dimension):
                                   eigen_vectors[:, ind, dimension]})

        dti_measures = DTIMeasures()
        fa, md = dti_measures.concat_and_calculate(eigen_values[:, 0], eigen_values[:, 1], eigen_values[:, 2])

        extra_maps.update({self.name + '.FA': fa,
                           self.name + '.MD': md,
                           self.name + '.AD': sorted_eigen_values[0],
                           self.name + '.RD': (sorted_eigen_values[1] + sorted_eigen_values[2]) / 2})

        return extra_maps



import numpy as np

from mdt.components_loader import bind_function
from mdt.models.single import DMRISingleModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Tensor(DMRISingleModelConfig):

    ex_vivo_suitable = False
    description = 'The standard Tensor model with in vivo defaults.'
    model_expression = '''
        S0 * Tensor
    '''
    inits = {'Tensor.d': 1.7e-9,
             'Tensor.dperp0': 1.7e-10,
             'Tensor.dperp1': 1.7e-10}

    @bind_function
    def _get_suitable_volume_indices(self, problem_data):
        protocol = problem_data.protocol
        unweighted_threshold = 25e6  # in SI units of s/m^2

        if protocol.has_column('g') and protocol.has_column('b'):
            protocol_indices = protocol.get_indices_bval_in_range(start=0, end=1.5e9 + 0.1e9)
            protocol_indices = np.append(protocol_indices, protocol.get_unweighted_indices(unweighted_threshold))
        else:
            protocol_indices = list(range(protocol.length))

        return np.unique(protocol_indices)


class TensorExVivo(Tensor):

    name = 'Tensor-ExVivo'
    in_vivo_suitable = False
    description = 'The standard Tensor model with ex vivo defaults.'
    inits = {'Tensor.d': 0.6e-9,
             'Tensor.dperp0': 0.6e-10,
             'Tensor.dperp1': 0.6e-10}


class TensorT2(Tensor):

    name = 'Tensor-T2'
    description = 'The Tensor model with in vivo defaults and extra T2 scaling.'
    model_expression = '''
            S0 * ExpT2Dec * Tensor
        '''


class TensorT2ExVivo(TensorExVivo):

    name = 'Tensor-ExVivo-T2'
    in_vivo_suitable = False
    description = 'The Tensor model with ex vivo defaults and extra T2 scaling.'
    model_expression = '''
            S0 * ExpT2Dec * Tensor
        '''



import glob
import os
from mdt.batch_utils import SimpleBatchProfile, BatchFitProtocolLoader, SimpleSubjectInfo

__author__ = 'Robbert Harms'
__date__ = "2015-07-13"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"

meta_info = {'title': 'Directory per subject',
             'description': 'General layout for batch fitting with a folder per subject.',
             'directory_layout':
'''
Every subject has its own directory. For every type of file (protocol, bvec, bval, TE, Delta, delta, DWI, mask)
we use the first one found.

You can provide the noise standard deviation to use using a noise_std file containing a single float.

Example directory layout:
    /a/*.nii(.gz)
    /a/*bval*
    /a/*bvec*

    /b/...
    /c/...

Optional items:
    /*/prtcl
    /*/*.prtcl

    /*/TE (in seconds)
    /*/delta (in seconds)
    /*/Delta (in seconds)

    /*/*_mask.nii(.gz)

    /*/noise_std.{txt,nii,nii.gz}

The optional items TE, Delta and delta provide extra information about the protocol.
They should either contain exactly 1 value (for all protocol lines), or a value per protocol line.
'''}


class DirPerSubject(SimpleBatchProfile):

    def __init__(self):
        super(DirPerSubject, self).__init__()
        self.use_gradient_deviations = False

    def _get_subjects(self):
        dirs = sorted([os.path.basename(f) for f in glob.glob(os.path.join(self._root_dir, '*'))])
        subjects = []

        for subject_id in dirs:
            niftis = glob.glob(os.path.join(self._root_dir, subject_id, '*.nii*'))
            dwis = list(filter(lambda v: all(name not in v for name in ['_mask', 'grad_dev', 'noise_std']), niftis))
            masks = list(filter(lambda v: '_mask' in v, niftis))
            grad_devs = list(filter(lambda v: 'grad_dev' in v, niftis))
            protocols = glob.glob(os.path.join(self._root_dir, subject_id, '*prtcl'))
            bvals = glob.glob(os.path.join(self._root_dir, subject_id, '*bval*'))
            bvecs = glob.glob(os.path.join(self._root_dir, subject_id, '*bvec*'))
            noise_std = self._autoload_noise_std(subject_id)

            if dwis:
                dwi_fname = dwis[0]

                mask_first_choice = glob.glob(os.path.join(self._root_dir, subject_id, 'mask.nii*'))
                if mask_first_choice:
                    mask_fname = mask_first_choice[0]
                else:
                    mask_fname = masks[0] if masks else None

                if not self.use_gradient_deviations:
                    grad_dev = None
                else:
                    grad_dev = grad_devs[0] if grad_devs else None

                protocol_fname = protocols[0] if protocols else None
                bval_fname = bvals[0] if bvals else None
                bvec_fname = bvecs[0] if bvecs else None

                if dwi_fname and (protocol_fname or (bval_fname and bvec_fname)):
                    protocol_loader = BatchFitProtocolLoader(
                        os.path.join(self._root_dir, subject_id),
                        protocol_fname=protocol_fname, bvec_fname=bvec_fname,
                        bval_fname=bval_fname)

                    output_dir = self._get_subject_output_dir(subject_id, mask_fname)

                    subjects.append(SimpleSubjectInfo(subject_id, dwi_fname, protocol_loader, mask_fname, output_dir,
                                                      gradient_deviations=grad_dev, noise_std=noise_std))

        return subjects

    def __str__(self):
        return meta_info['title']



import glob
import os
import numpy as np
import nibabel as nib
import scipy.io


__author__ = 'Robbert Harms'
__date__ = "2014-08-28"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Nifti(object):

    @staticmethod
    def write_volume_map(name, result_volume, directory, nifti_header, overwrite_volumes=True):
        """Write a single volume to the given directory.

        Args:
            name: the name of the volume
            result_volume: the volume we want to write out
            directory: the directory to write to
            nifti_header: the nifti header to use for each of the volumes
            overwrite_volumes: defaults to True, if we want to overwrite the volumes if they exists
        """
        Nifti.write_volume_maps({name: result_volume}, directory, nifti_header, overwrite_volumes)

    @staticmethod
    def write_volume_maps(result_volumes, directory, nifti_header, overwrite_volumes=True, gzip=True):
        """Write a number of maps (image result volumes) to the specific directory.

        Args:
            result_volumes (dict): the volume maps (3d) with the results we want to write out
                The naming of the file is the key of the volume with the extension appended by this function
            directory (str): the directory to write to
            nifti_header: the nifti header to use for each of the volumes
            overwrite_volumes (boolean): defaults to True, if we want to overwrite the volumes if they exists
            gzip (boolean): if True we write the files as .nii.gz, if False we write the files as .nii
        """
        if not os.path.exists(directory):
            os.makedirs(directory)

        for key, volume in result_volumes.items():
            extension = '.nii'
            if gzip:
                extension += '.gz'
            filename = key + extension

            full_filename = os.path.abspath(os.path.join(directory, filename))

            if os.path.exists(full_filename):
                if overwrite_volumes:
                    os.remove(full_filename)
                    nib.Nifti1Image(volume, None, nifti_header).to_filename(full_filename)
            else:
                nib.Nifti1Image(volume, None, nifti_header).to_filename(full_filename)

    @staticmethod
    def read_volume_maps(directory, map_names=None):
        """Read a number of Nifti volume maps that were written using write_volume_maps.

        Args:
            directory: the directory from which we want to read a number of maps
            map_names: the names of the maps we want to set_current_map. If given we only set_current_map and return these maps.

        Returns:
            A dictionary with the volumes. The keys of the dictionary are the filenames
            (without the extension) of the files in the given directory.
        """
        maps = {}
        for extension in ('.nii', '.nii.gz'):
            for f in glob.glob(os.path.join(directory, '*' + extension)):
                map_name = os.path.basename(f)[0:-len(extension)]

                if map_names is None or map_name in map_names:
                    maps.update({map_name: nib.load(f).get_data()})
        return maps

    @staticmethod
    def volume_names_generator(directory):
        """Get the names of the Nifti volume maps in the given directory.

        Args:
            directory: the directory to get the names of the available maps from

        Returns:
            generator: this yields the volume names in the given directory
        """
        for extension in ('.nii', '.nii.gz'):
            for f in glob.glob(os.path.join(directory, '*' + extension)):
                yield os.path.basename(f)[0:-len(extension)]


class TrackMark(object):
    """TrackMark is an proprietary visualization tool written by Alard Roebroeck and can be used to visualize fibre\
    directions.
    """

    @staticmethod
    def write_tvl_direction_pairs(tvl_filename, tvl_header, direction_pairs):
        """Write the given directions to TVL.

        The direction pairs should be a list with lists containing the vector and value to write. For example:
            ((vec, val), (vec1, val1), ...) up to three pairs are allowed.

        Args:
            tvl_filename (str): the filename to write to
            tvl_header (list): the header for the TVL file. This is a list of either 4 or 10 entries.
                4 entries: [version, res, gap, offset]
                10 entries: [version, x_res, x_gap, x_offset, y_res, y_gap, y_offset, z_res, z_gap, z_offset]
            direction_pairs (list of ndarrays): The list with direction pairs, only three are used.
                This is a list with (vector, magnitude) tuples in which the vectors are 4d volumes with for
                every voxel a 3d coordinate.
        """
        direction_pairs = direction_pairs[0:3]
        dir_matrix = np.zeros(direction_pairs[0][0].shape[0:3] + (12,))
        for ind, dirs in enumerate(direction_pairs):
            dir_matrix[..., ind*3:ind*3+3] = np.ascontiguousarray(np.squeeze(dirs[0]))
            dir_matrix[..., 9 + ind] = np.ascontiguousarray(np.squeeze(dirs[1]))

        TrackMark.write_tvl_matrix(tvl_filename, tvl_header, dir_matrix)

    @staticmethod
    def write_tvl_matrix(tvl_filename, tvl_header, directions_matrix):
        """Write the given directions matrix to TVL.

        Args:
            tvl_filename: the filename to write to
            tvl_header: the header for the TVL file. This is a list of either 4 or 10 entries.
                4 entries: [version, res, gap, offset]
                10 entries: [version, x_res, x_gap, x_offset, y_res, y_gap, y_offset, z_res, z_gap, z_offset]
            directions_matrix: an 4dimensional matrix, of which the fourth dimension is of length 12.
        """
        if os.path.exists(tvl_filename):
            os.remove(tvl_filename)

        if not os.path.exists(os.path.dirname(tvl_filename)):
            os.makedirs(os.path.dirname(tvl_filename))

        open(tvl_filename, 'a').close()
        with open(tvl_filename, 'rb+') as f:
            version = np.array(tvl_header[0]).astype(np.uint16)
            version.tofile(f, '')

            if len(tvl_header) == 4:
                for i in range(3):
                    np.array(directions_matrix.shape[i]).astype(np.uint32).tofile(f, '')
                    np.array(tvl_header[1]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[2]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[3]).astype(np.float64).tofile(f, '')
            else:
                for i in range(3):
                    np.array(directions_matrix.shape[i]).astype(np.uint32).tofile(f, '')
                    np.array(tvl_header[i * 3 + 1]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[i * 3 + 2]).astype(np.float64).tofile(f, '')
                    np.array(tvl_header[i * 3 + 3]).astype(np.float64).tofile(f, '')

            directions_matrix = np.transpose(directions_matrix, (3, 2, 1, 0)).astype(np.float32).flatten('F')
            directions_matrix.tofile(f, '')

    @staticmethod
    def write_rawmaps(directory, volumes, overwrite_volumes=True):
        """Write a dictionary with volumes to the given directory.

        Args:
            directory (str): the directory to write to
            volumes (dict): an dictionary with the volume maps (3d) with the results we want to write out
                The naming of the file is the key of the volume with .rawmap appended by this function.
            overwrite_volumes (boolean): if we want to overwrite already present volumes
        """
        if not os.path.isdir(directory):
            os.makedirs(directory)

        for key, volume in volumes.items():
            filename = key + '.rawmap'
            full_filename = os.path.abspath(os.path.join(directory, filename))

            if os.path.exists(full_filename):
                if overwrite_volumes:
                    os.remove(full_filename)
                    TrackMark.write_rawmap(full_filename, volume)
            else:
                TrackMark.write_rawmap(full_filename, volume)

    @staticmethod
    def write_rawmap(rawmap_filename, volume):
        """Write a rawmap to the given file.

        Args:
            rawmap_filename (str): The filename to write to, if not exists, it is created
                (along with extra directories). This should end on .rawmap, it not it is added.
            volume (ndarray): the volume to write. 3d or 4d. If 4d and 4th dimension is larger than 1
                additional maps are created.
        """
        if rawmap_filename[-len('.rawmap'):] != '.rawmap':
            rawmap_filename += '.rawmap'

        if os.path.exists(rawmap_filename):
            os.remove(rawmap_filename)

        if not os.path.exists(os.path.dirname(rawmap_filename)):
            os.makedirs(os.path.dirname(rawmap_filename))

        open(rawmap_filename, 'a').close()

        s = volume.shape
        if len(s) == 4:
            if s[3] == 1:
                volume = np.squeeze(volume, axis=(3,))
            else:
                subnames = rawmap_filename[0:-len('.rawmap')] + '_'
                for ind in range(volume.shape[3]):
                    TrackMark.write_rawmap(subnames + repr(ind) + '.rawmap', volume[..., ind])
                return

        with open(rawmap_filename, 'rb+') as f:
            np.array(s[0]).astype(np.uint16).tofile(f, '')
            np.array(s[1]).astype(np.uint16).tofile(f, '')
            np.array(s[2]).astype(np.uint16).tofile(f, '')

            m = np.transpose(volume, [2, 1, 0]).astype(np.float32).flatten('F')
            m.tofile(f, '')


class Matlab(object):

    @staticmethod
    def load_maps_from_mat(mat_file):
        """Load a result dictionary from the given matlab mat file

        Args:
            mat_file (str): The location of the .mat file.

        Returns:
            dict: All the items in the mat file by their name.
        """
        mat_data = scipy.io.loadmat(mat_file)
        fields = (e[0] for e in scipy.io.whosmat(mat_file))
        results = {}
        for field in fields:
            results.update({field: mat_data[field]})
        return results



import pyopencl as cl
import numpy as np
from mot.utils import get_float_type_def
from mot.cl_routines.base import CLRoutine
from mot.load_balance_strategies import Worker


__author__ = 'Robbert Harms'
__date__ = "2014-05-18"
__license__ = "LGPL v3"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class CalculateEigenvectors(CLRoutine):

    def convert_theta_phi_psi(self, theta_roi, phi_roi, psi_roi, double_precision=False):
        """Calculate the eigenvectors from the given theta, phi and psi angles.

        This will return the eigenvectors unsorted (since we know nothing about the eigenvalues).

        Args:
            theta_roi (ndarray): The list of theta's per voxel in the ROI
            phi_roi (ndarray): The list of phi's per voxel in the ROI
            psi_roi (ndarray): The list of psi's per voxel in the ROI
            double_precision (boolean): if we want to use float (set it to False) or double (set it to True)

        Returns:
            The three eigenvectors per voxel in the ROI. The return matrix is of shape (n, 3, 3) where n is the number
            of voxels and the second dimension holds the number of evecs and the last dimension the direction per evec.

            This gives for one voxel the matrix:
                [evec_1_x, evec_1_y, evec_1_z,
                 evec_2_x, evec_2_y, evec_2_z
                 evec_3_x, evec_3_y, evec_3_z]

            The resulting eigenvectors are the same as those from the Tensor.
        """
        np_dtype = np.float32
        if double_precision:
            np_dtype = np.float64

        theta_roi = np.require(theta_roi, np_dtype, requirements=['C', 'A', 'O'])
        phi_roi = np.require(phi_roi, np_dtype, requirements=['C', 'A', 'O'])
        psi_roi = np.require(psi_roi, np_dtype, requirements=['C', 'A', 'O'])

        rows = theta_roi.shape[0]
        evecs = np.zeros((rows, 3, 3), dtype=np_dtype, order='C')

        workers = self._create_workers(lambda cl_environment: _CEWorker(cl_environment, self.get_compile_flags_list(),
                                                                        theta_roi, phi_roi,
                                                                        psi_roi, evecs, double_precision))
        self.load_balancer.process(workers, rows)
        return evecs


class _CEWorker(Worker):

    def __init__(self, cl_environment, compile_flags, theta_roi, phi_roi, psi_roi, evecs, double_precision):
        super(_CEWorker, self).__init__(cl_environment)

        self._theta_roi = theta_roi
        self._phi_roi = phi_roi
        self._psi_roi = psi_roi
        self._evecs = evecs
        self._double_precision = double_precision
        self._all_buffers, self._evecs_buf = self._create_buffers()
        self._kernel = self._build_kernel(compile_flags)

    def calculate(self, range_start, range_end):
        nmr_problems = range_end - range_start

        event = self._kernel.generate_tensor(self._cl_run_context.queue, (int(nmr_problems), ), None, *self._all_buffers,
                                             global_offset=(int(range_start),))
        return [self._enqueue_readout(self._evecs_buf, self._evecs, range_start, range_end, [event])]

    def _create_buffers(self):
        thetas_buf = cl.Buffer(self._cl_run_context.context,
                               cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                               hostbuf=self._theta_roi)
        phis_buf = cl.Buffer(self._cl_run_context.context,
                             cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                             hostbuf=self._phi_roi)
        psis_buf = cl.Buffer(self._cl_run_context.context,
                             cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
                             hostbuf=self._psi_roi)
        evecs_buf = cl.Buffer(self._cl_run_context.context,
                              cl.mem_flags.WRITE_ONLY | cl.mem_flags.USE_HOST_PTR,
                              hostbuf=self._evecs)

        all_buffers = [thetas_buf, phis_buf, psis_buf, evecs_buf]
        return all_buffers, evecs_buf

    def _get_kernel_source(self):
        kernel_source = ''
        kernel_source += get_float_type_def(self._double_precision)
        kernel_source += '''
            __kernel void generate_tensor(
                global mot_float_type* thetas,
                global mot_float_type* phis,
                global mot_float_type* psis,
                global mot_float_type* evecs
                ){
                    int gid = get_global_id(0);

                    mot_float_type theta = thetas[gid];
                    mot_float_type phi = phis[gid];
                    mot_float_type psi = psis[gid];

                    mot_float_type cos_theta;
                    mot_float_type sin_theta = sincos(theta, &cos_theta);
                    mot_float_type cos_phi;
                    mot_float_type sin_phi = sincos(phi, &cos_phi);
                    mot_float_type cos_psi;
                    mot_float_type sin_psi = sincos(psi, &cos_psi);

                    mot_float_type4 n1 = (mot_float_type4)(cos_phi * sin_theta, sin_phi * sin_theta, cos_theta, 0.0);

                    // rotate around n1
                    mot_float_type tmp = sin(theta+(M_PI_2)); // using tmp as the rotation factor (90 degrees)
                    mot_float_type4 n2 = (mot_float_type4)(tmp * cos_phi, tmp * sin_phi, cos(theta+(M_PI_2)), 0.0);
                    tmp = select(1, -1, n1.z < 0 || ((n1.z == 0.0) && n1.x < 0.0)); // using tmp as the multiplier
                    n2 = n2 * cos_psi + (cross(n2, tmp * n1) * sin_psi) + (tmp * n1 * dot(tmp * n1, n2) * (1-cos_psi));

                    mot_float_type4 n3 = cross(n1, n2);

                    evecs[gid*9] = n1.x;
                    evecs[gid*9 + 1] = n1.y;
                    evecs[gid*9 + 2] = n1.z;

                    evecs[gid*9 + 3] = n2.x;
                    evecs[gid*9 + 4] = n2.y;
                    evecs[gid*9 + 5] = n2.z;

                    evecs[gid*9 + 6] = n3.x;
                    evecs[gid*9 + 7] = n3.y;
                    evecs[gid*9 + 8] = n3.z;
            }
        '''
        return kernel_source



from mdt.models.cascade import CascadeConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class Noddi(CascadeConfig):

    name = 'Noddi (Cascade)'
    description = 'Cascade for Noddi initialized from Ball&Stick.'
    models = ('BallStick (Cascade)',
              'Noddi')
    inits = {'Noddi': [('w_ic.w', lambda output_previous, output_all_previous: output_previous['w_stick.w'] / 2.0),
                       ('w_ec.w', lambda output_previous, output_all_previous: output_previous['w_stick.w'] / 2.0),
                       ('w_csf.w', 'w_ball.w'),
                       ('Noddi_IC.theta', 'Stick.theta'),
                       ('Noddi_IC.phi', 'Stick.phi')]}


class NoddiS0(CascadeConfig):

    name = 'Noddi (Cascade|S0)'
    description = 'Cascade for Noddi initialized with only an S0 fit.'
    models = ('S0',
              'Noddi')


class NoddiFixed(CascadeConfig):

    name = 'Noddi (Cascade|fixed)'
    description = 'Cascade for Noddi with fixed directions from Ball&Stick.'
    models = ('BallStick (Cascade)',
              'Noddi')
    inits = {'Noddi': [('w_ic.w', lambda output_previous, output_all_previous: output_previous['w_stick.w'] / 2.0),
                       ('w_ec.w', lambda output_previous, output_all_previous: output_previous['w_stick.w'] / 2.0),
                       ('w_csf.w', 'w_ball.w')]}
    fixes = {'Noddi': [('Noddi_IC.theta', 'Stick.theta'),
                       ('Noddi_IC.phi', 'Stick.phi')]}


class Noddi2(CascadeConfig):

    name = 'Noddi2 (Cascade)'
    description = 'Cascade for Noddi2 initialized from Ball & 2x Stick.'
    models = ('BallStickStick (Cascade)',
              'Noddi2')
    inits = {'Noddi': [('w_ic0.w', lambda output_previous, output_all_previous: output_previous['w_stick0.w'] / 2.0),
                       ('w_ec0.w', lambda output_previous, output_all_previous: output_previous['w_stick0.w'] / 2.0),
                       ('w_ic1.w', lambda output_previous, output_all_previous: output_previous['w_stick1.w'] / 2.0),
                       ('w_ec1.w', lambda output_previous, output_all_previous: output_previous['w_stick1.w'] / 2.0),
                       ('w_csf.w', 'w_ball.w'),
                       ('Noddi_IC0.theta', 'Stick0.theta'),
                       ('Noddi_IC0.phi', 'Stick0.phi'),
                       ('Noddi_IC1.theta', 'Stick1.theta'),
                       ('Noddi_IC1.phi', 'Stick1.phi')]}


class Noddi2Fixed(CascadeConfig):

    name = 'Noddi2 (Cascade|fixed)'
    description = 'Cascade for Noddi2 with fixed directions from Ball & 2x Stick.'
    models = ('BallStickStick (Cascade)',
              'Noddi2')
    inits = {'Noddi': [('w_ic0.w', lambda output_previous, output_all_previous: output_previous['w_stick0.w'] / 2.0),
                       ('w_ec0.w', lambda output_previous, output_all_previous: output_previous['w_stick0.w'] / 2.0),
                       ('w_ic1.w', lambda output_previous, output_all_previous: output_previous['w_stick1.w'] / 2.0),
                       ('w_ec1.w', lambda output_previous, output_all_previous: output_previous['w_stick1.w'] / 2.0),
                       ('w_csf.w', 'w_ball.w')]}
    fixes = {'Noddi': [('Noddi_IC0.theta', 'Stick0.theta'),
                       ('Noddi_IC0.phi', 'Stick0.phi'),
                       ('Noddi_IC1.theta', 'Stick1.theta'),
                       ('Noddi_IC1.phi', 'Stick1.phi')]}



from mdt.models.parameters import ProtocolParameterConfig

__author__ = 'Robbert Harms'
__date__ = "2015-12-12"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


"""The protocol parameters

The type of these parameters signifies that the data for this parameter should come from the protocol defined in the
model data. These will never be optimized and are always set to the data defined in the protocol.

Please choose the parameter type for a model and parameter carefully since the type signifies how the parameter and
its data are handled during model construction.

"""


class g(ProtocolParameterConfig):
    data_type = 'mot_float_type4'


class b(ProtocolParameterConfig):
    pass

class G(ProtocolParameterConfig):
    pass


class Delta(ProtocolParameterConfig):
    pass


class delta(ProtocolParameterConfig):
    pass


class q(ProtocolParameterConfig):
    pass


class TE(ProtocolParameterConfig):
    pass


class TM(ProtocolParameterConfig):
    pass


class Ti(ProtocolParameterConfig):
    pass


class TR(ProtocolParameterConfig):
    pass


class flip_angle(ProtocolParameterConfig):
    pass

#The next variables are for STEAM/TSE sequences, depending on the model in which they are used.
class Refoc_fa1(ProtocolParameterConfig):
    pass


class Refoc_fa2(ProtocolParameterConfig):
    pass


class SEf(ProtocolParameterConfig):
    pass



import numpy as np

from mdt.components_loader import bind_function
from mdt.models.single import DMRISingleModelConfig

__author__ = 'Robbert Harms'
__date__ = "2015-06-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


class S0(DMRISingleModelConfig):

    description = 'Models the unweighted signal (aka. b0).'
    model_expression = 'S0'

    @bind_function
    def _get_suitable_volume_indices(self, problem_data):
        protocol = problem_data.protocol
        protocol_indices = np.array([], dtype=np.int64)
        unweighted_threshold = 25e6 # in SI units of s/m^2

        if protocol.has_column('g') and protocol.has_column('b'):
            protocol_indices = np.append(protocol_indices, protocol.get_unweighted_indices(unweighted_threshold))
        else:
            protocol_indices = list(range(protocol.length))

        return np.unique(protocol_indices)
